"""Pytest configuration file."""

from datetime import datetime, timezone
from typing import AsyncGenerator

import pytest
import respx
from jose.jwt import encode

from vantage_agent.settings import SETTINGS


@pytest.fixture()
def token_content() -> str:
    """Generate a dummy token content."""
    one_minute_from_now = int(datetime.now(tz=timezone.utc).timestamp()) + 60
    return encode(dict(exp=one_minute_from_now), key="dummy-key", algorithm="HS256")


@pytest.fixture()
async def mock_access_token(token_content) -> AsyncGenerator[None, None]:
    """Fixture to mock the access token."""
    async with respx.mock:
        respx.post(f"https://{SETTINGS.OIDC_DOMAIN}/protocol/openid-connect/token").respond(
            status_code=200,
            json=dict(access_token=token_content),
        )
        yield



"""Core module for defining global settings of the Vantage Agent."""

import re
import sys
from pathlib import Path, PosixPath
from typing import Annotated, Optional

from pydantic import AnyHttpUrl, Field, confloat, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from vantage_agent.logger import logger


def _define_dotenv_file_location() -> PosixPath:
    """Define the location of the .env file based on the env var `DOTENV_FILE_LOCATION`.

    In case it is not set, the default location is defined as `.env`.
    """
    default_dotenv_file_location = Path("/var/snap/vantage-agent/common/.env")
    if default_dotenv_file_location.exists():
        return default_dotenv_file_location
    return Path(".env")


class Settings(BaseSettings):
    """Settings for the Vantage Agent."""

    PARTITIONS_JSON_PATH: Path = Path("/nfs/slurm/etc/aws/partitions.json")
    SCONTROL_PATH: Path = Path("/usr/bin/scontrol")
    SQUEUE_PATH: Path = Path("/usr/bin/squeue")
    SCANCEL_PATH: Path = Path("/usr/bin/scancel")
    SLURM_CONF_PATH: Path = Path("/etc/slurm/slurm.conf")
    GRES_CONF_PATH: Path = Path("/etc/slurm/gres.conf")
    SSSD_CONF_PATH: Path = Path("/etc/sssd/sssd.conf")

    IS_CLOUD_CLUSTER: bool = False

    # Vantage API info
    BASE_API_URL: Annotated[str, AnyHttpUrl] = "https://apis.vantagecompute.ai"

    # Sentry
    SENTRY_DSN: Optional[AnyHttpUrl] = None
    SENTRY_ENV: str = "local"
    SENTRY_TRACES_SAMPLE_RATE: Annotated[float, confloat(gt=0, le=1.0)] = 0.01
    SENTRY_SAMPLE_RATE: Annotated[float, confloat(gt=0.0, le=1.0)] = 0.25
    SENTRY_PROFILING_SAMPLE_RATE: Annotated[float, confloat(gt=0.0, le=1.0)] = 0.01

    # OIDC config for machine-to-machine security
    OIDC_DOMAIN: str
    OIDC_CLIENT_ID: str
    OIDC_CLIENT_SECRET: str
    OIDC_USE_HTTPS: bool = True
    CLUSTER_NAME: str

    CACHE_DIR: Path = Path.home() / ".cache/vantage-agent"

    # Task settings
    TASK_JOBS_INTERVAL_SECONDS: int = Field(30, ge=10, le=3600)  # seconds
    TASK_SELF_UPDATE_INTERVAL_SECONDS: Optional[int] = Field(None, ge=0)  # seconds

    @model_validator(mode="before")
    @classmethod
    def validade_cluster_name(cls, values):
        """Pre validator to set default values."""
        if not values.get("CLUSTER_NAME") and values.get("OIDC_CLIENT_ID"):
            client_regex = (
                r"^(.*)-([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})$"  # noqa
            )
            client_id_regex = re.compile(client_regex)
            match = client_id_regex.match(values.get("OIDC_CLIENT_ID"))
            if match is not None:
                cluster_name = match.group(1)
                values.update(
                    CLUSTER_NAME=cluster_name,
                )
        return values

    @field_validator("SCONTROL_PATH", mode="after")
    @classmethod
    def validate_scontrol_path(cls, v: Path) -> Path:
        """Ensure that the SCONTROL_PATH is an absolute path."""
        if not v.is_absolute():
            raise ValueError("SCONTROL_PATH must be an absolute path")
        return v

    model_config = SettingsConfigDict(
        env_prefix="VANTAGE_AGENT_", env_file=_define_dotenv_file_location(), extra="ignore"
    )


def _init_settings() -> Settings:
    try:
        return Settings()  # type: ignore[call-arg]
    except ValueError as e:
        logger.error(e)
        sys.exit(1)


SETTINGS = _init_settings()



"""Core module for testing the vantage_agent/tasks.py module."""

from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch

from vantage_agent.tasks import (
    cluster_config_task,
    cluster_node_task,
    cluster_partition_task,
    cluster_queue_actions_task,
    cluster_queue_task,
    report_health_status,
    self_update_agent,
    self_update_task,
    status_report_task,
    sync_cluster_partitions,
    sync_cluster_partitions_task,
    sync_cluster_teams,
    sync_cluster_teams_task,
    upload_scontrol_config,
    upload_scontrol_node,
    upload_scontrol_partition,
    upload_squeue_queue,
)


@pytest.fixture
def mock_scheduler() -> mock.MagicMock:
    """Mock the BaseScheduler class."""
    mock_scheduler = mock.Mock(spec=["add_job"])
    return mock_scheduler


@pytest.fixture
def mock_settings(monkeypatch: MonkeyPatch) -> mock.MagicMock:
    """Mock the settings module."""
    mock_settings = mock.Mock()
    monkeypatch.setattr("vantage_agent.tasks.SETTINGS", mock_settings)
    return mock_settings


class TestSchedulerTasks:
    """Test the task scheduling functions."""

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_cluster_config_task(
        self, interval: int | None, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the cluster_config_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        job = cluster_config_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(upload_scontrol_config, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_cluster_partition_task(
        self, interval: int | None, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the cluster_partition_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        job = cluster_partition_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(
            upload_scontrol_partition, "interval", seconds=interval
        )
        assert isinstance(job, mock.Mock)

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_cluster_queue_task(
        self, interval: int | None, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the cluster_queue_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        job = cluster_queue_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(upload_squeue_queue, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_cluster_node_task(
        self, interval: int | None, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the cluster_node_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        job = cluster_node_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(upload_scontrol_node, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_self_update_task_enabled(
        self, interval: int | None, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the self_update_task function with a valid interval."""
        mock_settings.TASK_SELF_UPDATE_INTERVAL_SECONDS = interval
        job = self_update_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(self_update_agent, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    def test_self_update_task_disabled(self, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock):
        """Test the self_update_task function with a None interval."""
        mock_settings.TASK_SELF_UPDATE_INTERVAL_SECONDS = None
        job = self_update_task(mock_scheduler)
        mock_scheduler.add_job.assert_not_called()
        assert job is None

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_status_report_task(
        self, interval: int, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the status_report_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        job = status_report_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(report_health_status, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_sync_slurm_partitions_task(
        self, interval: int, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the sync_cluster_partitions_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        job = sync_cluster_partitions_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(sync_cluster_partitions, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    def test_sync_slurm_partitions_task_disabled(
        self, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test if the `sync_cluster_partitions` task is not scheduled for on-premises clusters."""
        mock_settings.IS_CLOUD_CLUSTER = False
        job = sync_cluster_partitions_task(mock_scheduler)
        mock_scheduler.add_job.assert_not_called()
        assert job is None

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_cluster_queue_actions_task(
        self, interval: int | None, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the cluster_queue_actions_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        from vantage_agent.logicals.queue_actions import process_cluster_queue_actions

        job = cluster_queue_actions_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(
            process_cluster_queue_actions, "interval", seconds=interval
        )
        assert isinstance(job, mock.Mock)

    @pytest.mark.parametrize("interval", [30, 60, 100])
    def test_sync_cluster_teams_task(
        self, interval: int, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Test the sync_cluster_teams_task function."""
        mock_settings.TASK_JOBS_INTERVAL_SECONDS = interval
        mock_settings.IS_CLOUD_CLUSTER = False
        job = sync_cluster_teams_task(mock_scheduler)
        mock_scheduler.add_job.assert_called_once_with(sync_cluster_teams, "interval", seconds=interval)
        assert isinstance(job, mock.Mock)

    def test_sync_cluster_teams_task_cloud_clusters(
        self, mock_scheduler: mock.MagicMock, mock_settings: mock.MagicMock
    ):
        """Ensure the sync_cluster_teams task is skipped for cloud clusters."""
        mock_settings.IS_CLOUD_CLUSTER = True

        job = sync_cluster_teams_task(mock_scheduler)

        mock_scheduler.add_job.assert_not_called()
        assert job is None



"""Core module for testing the vantage_agent/logicals/scontrol.py module."""

import asyncio
import subprocess
from typing import Generator
from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from jsondiff import diff

from vantage_agent.logicals.squeue import (
    SETTINGS,
    upload_squeue_queue,
)


@pytest.fixture
def mock_upload_squeue_queue(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the upload_slurm_queue function."""
    mock_upload = mock.AsyncMock()
    monkeypatch.setattr("vantage_agent.logicals.squeue.upload_slurm_queue", mock_upload)
    return mock_upload


@pytest.fixture
def mock_subprocess_run(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the subprocess.run function."""
    mock_run = mock.Mock()
    mock_run.return_value.stdout = "Sample squeue config output"
    monkeypatch.setattr(subprocess, "run", mock_run)
    yield mock_run


@pytest.fixture
def mock_load_cached_dict(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the load_cached_dict function."""
    mock_load = mock.Mock()
    mock_load.return_value = [{"sydney": "perth"}]
    monkeypatch.setattr("vantage_agent.logicals.squeue.load_cached_dict", mock_load)
    return mock_load


@pytest.fixture
def mock_cache_dict(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the load_cached_dict function."""
    mock_cache = mock.Mock()
    monkeypatch.setattr("vantage_agent.logicals.squeue.cache_dict", mock_cache)
    return mock_cache


@pytest.fixture
def mock_asyncio_run(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the asyncio.run function."""
    mock_run = mock.MagicMock()
    monkeypatch.setattr(asyncio, "run", mock_run)
    return mock_run


@pytest.fixture
def mock_logger(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the logger."""
    mock_logger = mock.Mock()
    monkeypatch.setattr("vantage_agent.logicals.squeue.logger", mock_logger)
    return mock_logger


@pytest.fixture
def mock_parse_slurm_queue(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the parse_slurm_queue function."""
    mock_parse = mock.Mock()
    mock_parse.return_value = [{"state": "RUNNING", "name": "job-name"}]
    monkeypatch.setattr("vantage_agent.logicals.squeue.parse_slurm_queue", mock_parse)
    return mock_parse


class TestUploadSqueueQueue:
    """Test cases for the upload_squeue_queue function."""

    def test_queue_changes_detected(
        self,
        mock_subprocess_run: mock.MagicMock,
        mock_parse_slurm_queue: mock.MagicMock,
        mock_cache_dict: mock.MagicMock,
        mock_load_cached_dict: mock.MagicMock,
        mock_asyncio_run: mock.MagicMock,
        mock_logger: mock.MagicMock,
        mock_upload_squeue_queue: mock.MagicMock,
    ):
        """Test when there are changes in the Slurm queue."""
        upload_squeue_queue()

        mock_subprocess_run.assert_called_once_with(
            [SETTINGS.SQUEUE_PATH, "-o", "%all"], stdout=subprocess.PIPE, text=True
        )
        mock_parse_slurm_queue.assert_called_once()
        mock_load_cached_dict.assert_called_once_with("slurm_queue.json")
        mock_upload_squeue_queue.assert_called_once_with(
            diff(mock_load_cached_dict.return_value, mock_parse_slurm_queue.return_value, marshal=True)
        )
        mock_logger.debug.assert_has_calls(
            calls=[
                mock.call("Getting Slurm's queue information"),
                mock.call("Parsing Slurm's queue"),
                mock.call("Loading cached Slurm's queue"),
                mock.call("Computing difference between cached and current queue"),
                mock.call("Detected changes in Slurm's configuration. Uploading to the API..."),
            ]
        )
        mock_cache_dict.assert_called_once_with(mock_parse_slurm_queue.return_value, "slurm_queue.json")
        assert mock_asyncio_run.call_count == 1

    def test_no_queue_changes(
        self,
        mock_subprocess_run: mock.MagicMock,
        mock_parse_slurm_queue: mock.MagicMock,
        mock_load_cached_dict: mock.MagicMock,
        mock_asyncio_run: mock.MagicMock,
        mock_logger: mock.MagicMock,
        mock_upload_squeue_queue: mock.MagicMock,
    ):
        """Test when there are no changes in the Slurm queue."""
        mock_parse_slurm_queue.return_value = mock_load_cached_dict.return_value

        upload_squeue_queue()

        mock_subprocess_run.assert_called_once_with(
            [SETTINGS.SQUEUE_PATH, "-o", "%all"],
            stdout=subprocess.PIPE,
            text=True,  # noqa
        )
        mock_parse_slurm_queue.assert_called_once()
        mock_load_cached_dict.assert_called_once_with("slurm_queue.json")
        mock_asyncio_run.assert_not_called()
        mock_upload_squeue_queue.assert_not_called()
        mock_logger.debug.assert_has_calls(
            calls=[
                mock.call("Getting Slurm's queue information"),
                mock.call("Parsing Slurm's queue"),
                mock.call("Loading cached Slurm's queue"),
                mock.call("Computing difference between cached and current queue"),
                mock.call("No changes detected in Slurm's queue."),
            ]
        )



"""Test the plugin module.

Copied and adapted from https://github.com/omnivector-solutions/jobbergate/blob/main/jobbergate-agent/tests/utils.py/test_plugin.py.
"""

from unittest import mock

import pytest

from vantage_agent.plugin import load_plugins
from vantage_agent.tasks import (
    cluster_config_task,
    cluster_node_task,
    cluster_partition_task,
    cluster_queue_actions_task,
    cluster_queue_task,
    self_update_task,
    status_report_task,
    sync_cluster_partitions_task,
    sync_cluster_teams_task,
)


def test_discover_tasks__success():
    """Test that discover_tasks returns the expected result."""
    expected_result = {
        "cluster-config": cluster_config_task,
        "cluster-partition": cluster_partition_task,
        "cluster-queue": cluster_queue_task,
        "cluster-queue-actions": cluster_queue_actions_task,
        "cluster-node": cluster_node_task,
        "partitions-sync": sync_cluster_partitions_task,
        "teams-sync": sync_cluster_teams_task,
        "self-update": self_update_task,
        "status-report": status_report_task,
    }
    actual_result = load_plugins("tasks")

    assert actual_result == expected_result


@mock.patch("vantage_agent.plugin.entry_points")
def test_discover__fail_to_load(mocked_entry_points):
    """Test that discover_tasks raises RuntimeError when fails to load."""
    mocked_pluging = mock.Mock()
    mocked_pluging.load.side_effect = Exception("Test")
    mocked_entry_points.return_value = [mocked_pluging]

    with pytest.raises(RuntimeError, match="^Failed to load plugin"):
        load_plugins("non-existent-plugin")



"""Core module for defining operations related to cluster queue actions."""

import asyncio
import subprocess
from string import Template
from textwrap import dedent
from typing import Any, TypedDict

from loguru import logger

from vantage_agent.exceptions import VantageApiError
from vantage_agent.logger import log_error
from vantage_agent.settings import SETTINGS
from vantage_agent.vantage_api_client import AsyncBackendClient


class QueueActionResponse(TypedDict):
    """Type definition for the queue action response."""

    data: dict[Any, Any]
    errors: list[Any] | None


async def fetch_cluster_queue_actions() -> list[dict[str, Any]]:
    """Fetch all cluster queue actions for the current cluster."""
    logger.debug("Fetching cluster queue actions from Vantage API")

    query = dedent(
        Template(
            """
        query clusterQueueActions {
            clusterQueueActions(
                first: 100
                filters: {
                    clusterName: {
                        eq: "$clusterName"
                    }
                }
            ) {
                edges {
                    node {
                        id
                        clusterName
                        queueId
                        action
                        queue {
                            id
                            name
                            info
                        }
                    }
                }
                total
            }
        }
        """
        ).substitute(clusterName=SETTINGS.CLUSTER_NAME)
    )

    body = {"query": query}

    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: QueueActionResponse = res.json()

    if response_data.get("errors"):
        logger.error(f"Error fetching cluster queue actions: {response_data['errors']}")
        return []

    cluster_actions = response_data.get("data", {}).get("clusterQueueActions", {}).get("edges", [])
    logger.debug(f"Found {len(cluster_actions)} queue actions for cluster {SETTINGS.CLUSTER_NAME}")
    return cluster_actions


async def remove_queue_action(action_id: int) -> bool:
    """Remove a queue action from the database."""
    logger.debug(f"Removing queue action with ID: {action_id}")

    query = dedent(
        """
        mutation removeQueueAction($id: Int!) {
            removeQueueAction(id: $id) {
                __typename
                ... on RemoveQueueActionSuccess {
                    message
                }
                ... on InvalidInput {
                    message
                }
            }
        }
        """
    )

    body = {"query": query, "variables": {"id": action_id}}

    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: QueueActionResponse = res.json()

    if response_data.get("errors"):
        logger.error(f"Error removing queue action {action_id}: {response_data['errors']}")
        return False

    result = response_data.get("data", {}).get("removeQueueAction", {})
    if result.get("__typename") == "RemoveQueueActionSuccess":
        logger.debug(f"Successfully removed queue action {action_id}")
        return True
    elif result.get("__typename") == "InvalidInput":
        logger.warning(f"Queue action {action_id} not found: {result.get('message')}")
        return True  # Consider it successful since the action is already gone
    else:
        logger.error(f"Unexpected response when removing queue action {action_id}: {result}")
        return False


def execute_scancel(job_id: str) -> bool:
    """Execute scancel command for a specific job ID."""
    logger.debug(f"Executing scancel for job ID: {job_id}")

    try:
        result = subprocess.run(
            [SETTINGS.SCANCEL_PATH, job_id],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=2,
        )

        if result.returncode == 0:
            logger.info(f"Successfully cancelled job {job_id}")
            return True
        else:
            stderr_output = result.stderr.lower()
            if "invalid job id" in stderr_output or "no such job" in stderr_output:
                logger.warning(f"Job {job_id} does not exist or was already cancelled")
                return True
            else:
                logger.error(f"Failed to cancel job {job_id}: {result.stderr}")
                return False

    except subprocess.TimeoutExpired:
        logger.error(f"Timeout while trying to cancel job {job_id}")
        return False
    except Exception as e:
        logger.error(f"Exception while trying to cancel job {job_id}: {e}")
        return False


def extract_job_id_from_queue_info(queue_info: dict[str, Any]) -> str | None:
    """Extract job ID from queue info dictionary."""
    if not queue_info:
        return None

    if isinstance(queue_info, dict):
        if "jobid" in queue_info:
            return str(queue_info["jobid"])
        else:
            logger.warning(f"Could not extract job ID from queue info: {queue_info}")
            return None


def process_cluster_queue_actions():
    """Process all cluster queue actions - cancel jobs and clean up actions."""
    logger.debug("Processing cluster queue actions")

    with VantageApiError.handle_errors("Failed to process cluster queue actions", do_except=log_error):
        asyncio.run(_process_cluster_queue_actions_async())


async def _process_cluster_queue_actions_async():
    """Async implementation of process_cluster_queue_actions."""
    actions = await fetch_cluster_queue_actions()

    if not actions:
        logger.debug("No queue actions to process")
        return

    for action in actions:
        action_id = action["node"]["id"]
        queue_info = action["node"].get("queue", {}).get("info", {})
        action_type = action["node"]["action"]

        logger.debug(f"Processing action {action_id} of type {action_type}")

        if action_type.lower() != "cancel":
            logger.warning(f"Unsupported action type: {action_type}")
            continue

        job_id = extract_job_id_from_queue_info(queue_info)

        if job_id is None:
            logger.error(f"Could not extract job ID for action {action_id}, removing action")
            await remove_queue_action(action_id)
            continue

        success = execute_scancel(job_id)

        if success:
            logger.info(f"Successfully cancelled job {job_id} for action {action_id}")
        else:
            logger.error(f"Failed to cancel job {job_id} for action {action_id}")



"""Core module for testing the vantage_agent/scheduler.py module.

Copied and adapted from https://github.com/omnivector-solutions/jobbergate/blob/main/jobbergate-agent/tests/utils.py/test_scheduler.py.
"""

from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch

from vantage_agent.scheduler import (
    AsyncIOScheduler,
    init_scheduler,
    load_plugins,
    schedule_tasks,
    shut_down_scheduler,
)


@pytest.fixture
def mock_load_plugins(monkeypatch: MonkeyPatch) -> mock.MagicMock:
    """Mock the load_plugins function."""
    mock_scheduler = mock.Mock()
    monkeypatch.setattr("vantage_agent.scheduler.load_plugins", mock_scheduler)
    return mock_scheduler


@pytest.fixture
def mock_scheduler(monkeypatch: MonkeyPatch) -> mock.MagicMock:
    """Mock the AsyncIOScheduler class."""
    mock_scheduler = mock.Mock()
    monkeypatch.setattr("vantage_agent.scheduler.scheduler", mock_scheduler)
    return mock_scheduler


class TestScheduleTasks:
    """Test the schedule_tasks function."""

    def test_schedule_tasks__success(self, mock_load_plugins: mock.MagicMock):
        """Test that schedule_tasks returns the expected result."""
        discovered_functions = load_plugins("tasks")
        mocked_functions = {name: mock.Mock() for name in discovered_functions.keys()}

        mock_load_plugins.return_value = mocked_functions

        scheduler = AsyncIOScheduler()
        schedule_tasks(scheduler)

        mock_load_plugins.assert_called_once_with("tasks")
        assert all(m.call_count == 1 for m in mocked_functions.values())
        assert all(m.call_args(scheduler=scheduler) for m in mocked_functions.values())

    def test_schedule_tasks__fails(self, mock_load_plugins: mock.MagicMock):
        """Test that schedule_tasks raises RuntimeError when fails to schedule a task."""
        mocked_functions = {"supposed-to-fail": mock.Mock(side_effect=Exception("Test"))}

        mock_load_plugins.return_value = mocked_functions

        scheduler = AsyncIOScheduler()

        with pytest.raises(RuntimeError, match="^Failed to execute"):
            schedule_tasks(scheduler)

        mock_load_plugins.assert_called_once_with("tasks")


@mock.patch("vantage_agent.scheduler.schedule_tasks")
def test_init_scheduler(mock_schedule_tasks: mock.MagicMock, mock_scheduler: mock.MagicMock):
    """Test the init_scheduler function."""
    scheduler = init_scheduler()
    mock_scheduler.start.assert_called_once_with()
    mock_schedule_tasks.assert_called_once_with(mock_scheduler)
    assert scheduler is mock_scheduler


@pytest.mark.parametrize("wait", [True, False])
def test_shut_down_scheduler(wait: bool, mock_scheduler: mock.MagicMock):
    """Test the shut_down_scheduler function."""
    shut_down_scheduler(mock_scheduler, wait)
    mock_scheduler.shutdown.assert_called_once_with(wait)



"""Core module for testing the vantage_agent/logger.py module."""

from traceback import format_tb
from unittest import mock

import pytest

from vantage_agent.logger import DoExceptParams, log_error, logger_wraps


def sample_function(a, b):
    """This is a sample function to be decorated for testing."""  # noqa: D401,D404
    return a + b


@pytest.fixture
def mock_logger(monkeypatch):
    """Mock the logger object."""
    mock_logger = mock.Mock()
    monkeypatch.setattr("vantage_agent.logger.logger", mock_logger)
    return mock_logger


class TestLogError:
    """Test the log_error function."""

    def test_log_error(self, mock_logger):
        """Test the log_error function."""
        try:
            1 / 0
        except ZeroDivisionError as e:
            params = DoExceptParams(
                err=e,
                base_message="Sample error message",
                final_message="Sample error message",
                trace=e.__traceback__,
            )
            log_error(params)

        expected_message = "\n".join(
            [
                "Sample error message",
                "--------",
                "Traceback:",
                "".join(format_tb(params.trace)),
            ]
        )
        mock_logger.error.assert_called_once_with(expected_message)


class TestLoggerWraps:
    """Test the logger_wraps decorator."""

    def test_entry_only(self, mock_logger):
        """Test logging only the entry message."""

        @logger_wraps(exit=False)
        def decorated_func(a, b):
            return sample_function(a, b)

        result = decorated_func(2, 3)

        mock_logger.opt.assert_called_once_with(depth=1)
        mock_logger.opt.return_value.log.assert_called_once_with(
            "DEBUG", "Entering 'decorated_func' (args=(2, 3), kwargs={})"
        )
        assert result == 5

    def test_exit_only(self, mock_logger):
        """Test logging only the exit message."""

        @logger_wraps(entry=False)
        def decorated_func(a, b):
            return sample_function(a, b)

        result = decorated_func(2, 3)

        mock_logger.opt.assert_called_once_with(depth=1)
        mock_logger.opt.return_value.log.assert_called_once_with(
            "DEBUG", "Exiting 'decorated_func' (result=5)"
        )
        assert result == 5

    def test_no_logging(self, mock_logger):
        """Test that no logging occurs when both entry and exit are False."""

        @logger_wraps(entry=False, exit=False)
        def decorated_func(a, b):
            return sample_function(a, b)

        result = decorated_func(2, 3)

        mock_logger.opt.assert_called_once_with(depth=1)
        mock_logger.opt.return_value.log.assert_not_called()
        assert result == 5

    def test_custom_log_level(self, mock_logger):
        """Test logging at a custom log level (INFO)."""

        @logger_wraps(level="INFO")
        def decorated_func(a, b):
            return sample_function(a, b)

        result = decorated_func(2, 3)

        mock_logger.opt.assert_called_once_with(depth=1)
        mock_logger.opt.return_value.log.assert_has_calls(
            [
                mock.call("INFO", "Entering 'decorated_func' (args=(2, 3), kwargs={})"),
                mock.call("INFO", "Exiting 'decorated_func' (result=5)"),
            ]
        )
        assert result == 5



"""Core module for logging operations."""

import functools
from traceback import format_tb

from buzz import DoExceptParams
from loguru import logger


def log_error(params: DoExceptParams):
    """Log a Buzz-based exception and the stack-trace of the error's context.

    :param: params: A DoExceptParams instance containing the original exception, a
                    message describing it, and the stack trace of the error.
    """
    logger.error(
        "\n".join(
            [
                params.final_message,
                "--------",
                "Traceback:",
                "".join(format_tb(params.trace)),
            ]
        )
    )


def logger_wraps(*, entry: bool = True, exit: bool = True, level: str = "DEBUG"):
    """Wrap a function with logging statements.

    Reference:
        https://loguru.readthedocs.io/en/stable/resources/recipes.html
    """

    def wrapper(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            logger_ = logger.opt(depth=1)
            if entry:
                logger_.log(level, f"Entering '{name}' (args={args}, kwargs={kwargs})")
            result = func(*args, **kwargs)
            if exit:
                logger_.log(level, f"Exiting '{name}' (result={result})")
            return result

        return wrapped

    return wrapper
